from concurrent.futures import process
from re import L
from tkinter import Image
from typing import List, Dict, Union, Optional

from discord import Embed as DPYEMBED
from discord import Member, User, ClientUser, Color

from .object import *
from .exceptions import *

ANY_USER = Union[User, Member, ClientUser]

class Embed(DPYEMBED):
    def __init__(
        self,
        embed_type: Optional[EmbedType] = EmbedType.Rich,
        title: Optional[str] = None,
        url: Optional[str] = None,
        description: Optional[str] = "",
        color: Optional[Color] = Color.blurple(),
        timestamp: Optional[datetime] = None,
        author: Optional[Union[AuthorObject, Dict[str, str], None]] = None,
        footer: Optional[Union[FooterObject, Dict[str, str], None]] = None,
        thumbnail: Optional[Union[ThumbnailObject, Dict[str, Union[str, int]], None]] = None,
        image: Optional[Union[ThumbnailObject, Dict[str, Union[str, int]], None]] = None,
        provider: Optional[Union[ProviderObject, Dict[str, Any]]] = None,
        fields: Optional[Union[Fields, List[Field]]] = None
    ):
        self._type: EmbedType = embed_type if embed_type in EmbedType else EmbedType.from_valye(embed_type)
        self._title: str = process_title(title)
        self._url: str = url if validate_url(url) else None
        self._description: str = process_desc(description)
        
        if isinstance(color, Color):
            self._color = color
        elif isinstance(color, str):
            color = getattr(Color, color, None)
            if isinstance(color, classmethod):
                self._color = color()
            else:
                self._color = Color.blurple()
                raise ValueError("Invalid color key is passed.")
            
        self._timestamp: datetime = timestamp if type(timestamp) == datetime else None
        self._author: AuthorObject = author if isinstance(author, AuthorObject) else AuthorObject.fromDict(author)
        self._footer: FooterObject = footer if isinstance(footer, FooterObject) else FooterObject.fromDict(footer)
        self._thumbnail: ImageObject = thumbnail if isinstance(thumbnail, ImageObject) else ImageObject.fromDict(thumbnail)
        self._image: ImageObject = image if isinstance(image, ImageObject) else ImageObject.fromDict(image)
        self._provider: ProviderObject = provider if isinstance(provider, ProviderObject) else ProviderObject.fromDict(provider)
        self._fields: Fields = Fields.fromDict(fields)
        
    @property
    def title(self) -> str:
        return self._title
    
    @title.setter
    def title(self, value: str) -> NoReturn:
        self._title = process_title(value)
        
    @property
    def type(self) -> EmbedType:
        return self._type

    @type.setter
    def type(self, value: str) -> NoReturn:
        self._type = EmbedType.from_value(value)

    @property
    def description(self) -> str:
        return self._description

    @description.setter
    def description(self, value: str) -> NoReturn:
        if check_desc(value):
            self._description = value

    @property
    def color(self) -> Color:
        return self._color

    @color.setter
    def color(self, value: Color) -> NoReturn:
        if isinstance(value, Color):
            self._color = value
        else:
            self._color = process_color(value) # type: ignore

    @property
    def author(self) -> AuthorObject:
        return self._author

    @author.setter
    def author(self, value: AuthorObject) -> NoReturn:
        self._author = AuthorObject.fromDict(value)

    @property
    def footer(self) -> Dict[str, str]:
        return self._footer

    @footer.setter
    def footer(self, value: Dict[str, str]) -> NoReturn:
        self._footer = FooterObject.fromDict(value)

    @property
    def timestamp(self) -> datetime:
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value) -> NoReturn:
        if isinstance(value, datetime):
            self._timestamp = value
        else:
            raise TypeError("Timestamp object must be an instance of datetime")

    @property
    def url(self) -> str:
        return self._url

    @url.setter
    def url(self, value) -> NoReturn:
        if validate_url(value):
            self._url = value

    @property
    def thumbnail(self) -> ImageObject:
        return self._thumbnail

    @thumbnail.setter
    def thumbnail(self, value: Union[ImageObject, str]) -> NoReturn:
        try:
            self._thumbnail = ImageObject.fromDict(value)
        except:
            if validate_url(value):
                self._thumbnail = ImageObject(url=value)
            else:
                raise

    @property
    def image(self) -> ImageObject:
        return self._image

    @image.setter
    def image(self, value: Union[ImageObject, str]) -> NoReturn:
        try:
            self._image = ImageObject.fromDict(value)
        except:
            if validate_url(value):
                self._image = ImageObject(url=value)
            else:
                raise

    @property
    def fields(self) -> List[Field]:
        return self._fields

    @fields.setter
    def fields(self, value: List[Field]) -> NoReturn:
        self._fields = Fields.fromDict(value)

        if check_fields(value):  # type: ignore
            self._fields.extend(value)


import pkg_resources
from typing import NamedTuple, Literal

__all__ = (
    "__path__",
    "__author__",
    "__copyright__",
    "__license__",
    "__title__",
    "__version__",
    "version_info",
    "VersionInfo",
)

class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: Literal["alpha", "beta", "candidate", "final"]
    serial: int
    
version_info: VersionInfo = VersionInfo(major=0, minor=1, micro=3, releaselevel="alpha", serial=0)

__path__ = __import__("pkgutil").extend_path(__path__, __name__)
__author__ = 'Lemon Rose (japandotorg)'
__copyright__ = 'Copyright 2022 Lemon Rose (japandotorg) Melon-Development'
__license__ = 'MIT'
__title__ = 'melonutils'
__version__ = '.'.join(map(str, (version_info.major, version_info.minor, version_info.micro)))

pkg_resources.working_set.by_key.pop("melonutils", None)



from __future__ import annotations

import asyncio
from time import time_lib
from datetime import datetime
from typing import TYPE_CHECKING, TypeVar, Callable, Awaitable, Union, Any, Optional, Tuple

import discord
from redbot.core import commands # type: ignore

try:
    from typing import ParamSpec # type: ignore
except ImportError:
    from typing_extensions import ParamSpec
    
from .exceptions import *
    
T = TypeVar("T")
P = ParamSpec("P")

__all__: Tuple[str, ...] = (
    "ascii_color",
    "markdown_remove",
    "codeblock_wrapper",
    "safe_reason",
    "format_date",
    "add_logging",
    "can_execute_action",
)

def ascii_color(color=None, /, *, fmt=0, bg=False) -> str:
    """
    Returns the ascii color escape string for the given number.
    
    :param color: The color number.
    :param fmt: The format number.
    :param bg: Whether to return as a background color
    """
    base = "\u001b["
    
    if fmt != 0:
        base += "{fmt};"
    
    if color is None:
        base += "{color}m"
        color = 0
    
    else:
        if bg is True:
            base += "4{color}m"
        
        else:
            base += "3{color}m"
    
    return base.format(fmt=fmt, color=color)

def markdown_remove(entity: Any) -> str:
    """
    Returns the string of an object with discord markdown removed.
    
    Parameters
    ----------
    entity: Any
        The object to remove markdown from.
    
    Returns
    -------
    str
        The string of the object with markdown removed.
    """
    return discord.utils.remove_markdown(discord.utils.escape_mentions(str(entity)))

def codeblock_wrapper(text: str, /, *, lang: str = "py"):
    """
    Wraps a string into a code-block, and adds zero width
    characters to avoid the code block getting cut off.
    
    Parameters
    ----------
    text: str
        The text to wrap.
    lang: str
        The code language to use.
    
    Returns
    -------
    str
        The wrapped text.
    """
    text = text.replace("`", "\u200b`")
    return f"```{lang}\n{text}\n```"

def safe_reason(
    author: Union[discord.Member, discord.User], 
    reason: str, 
    *, 
    length: int = 512
) -> str:
    base = f"Action by {author} ({author.id}) for: "
    
    length_limit = length - len(base)
    
    if len(reason) > length_limit:
        reason = reason[: length_limit - 3] + "..."
        
    return base + reason

def format_date(
    date: datetime
) -> str:
    """
    Formats a date to a string in the preferred way.
    
    Parameters
    ----------
    date: datetime.datetime
        The date to format.
    
    Returns
    -------
    str
        The formatted date.
    """
    return date.strftime("%b %d, %Y %H:%M %Z")

def add_logging(
    func: Callable[P, Union[Awaitable[T], T]] # type: ignore
) -> Callable[P, Union[Awaitable[T], T]]: # type: ignore
    """
    Adds logging to a coroutine or a function.
    
    .. code-block:: python3
    
        >>> async def foo(a: int, b: int) -> int:
        >>>     return a + b
    
        >>> logger = add_logging(foo)
        >>> result = await logger(1, 2)
        >>> print(result)
        3
    
        >>> def foo(a: int, b: int) -> int:
        >>>     return a + b
    
        >>> logger = add_logging(foo)
        >>> result = logger(1, 2)
        >>> print(result)
        3
    """
    
    async def _async_wrapped(
        *args: P.args,
        **kwargs: P.kwargs
    ) -> Awaitable[T]:
        start = time_lib.time()
        
        result = await func(*args, **kwargs)
        
        print(f"{func.__name__} took {time_lib.time() - start:.2f} seconds")
        
        return result
    
    def _sync_wrapped(
        *args: P.args,
        **kwargs: P.kwargs
    ) -> T:
        start = time_lib.time()
        
        result = func(*args, **kwargs)
        
        print(f'{func.__name__} took {time_lib.time() - start:.2f} seconds')
        
        return result
    
    return _async_wrapped if asyncio.iscoroutinefunction(func) else _sync_wrapped # type: ignore

async def can_execute_action(
    ctx: commands.Context,
    target: Union[discord.Member, discord.User],
    *,
    fail_if_not_upgrade: bool = True,
) -> Optional[bool]:
    """
    |coro|
    
    A wrapped predicate to check if the action can be executed.
    
    Parameters
    ----------
    ctx: :class:`commands.Context`
        The context of the command.
    target: Union[:class:`discord.Member`, :class:`discord.User`]
        The target of the action.
    fail_if_not_upgrade: :class:`bool`
        Whether to fail if the user can't be upgraded to a Member.
    
    Returns
    -------
    Optional[:class:`bool`]
        Whether the action can be executed.
    
    Raises
    ------
    HierarchyException
        The action cannot be executed due to role hierarchy.
    ActionNotExecutable
        The action cannot be executed due to other reasons.
    commands.NoPrivateMessage
        This command cannot be used in private messages.
    """
    guild = ctx.guild
    
    if guild is None or not isinstance(ctx.author, discord.Member):
        raise commands.NoPrivateMessage('This command cannot be used in private messages.')
    
    if isinstance(target, discord.User):
        upgraded = await ctx.bot.get_or_fetch_member(guild, target)
        
        if upgraded is None:
            if fail_if_not_upgrade:
                raise ActionNotExecutable("That user is not a member of this server.")
            
        else:
            target = upgraded
            
    if ctx.author == target:
        raise ActionNotExecutable("You cannot execute this action on yourself!")
    
    if guild.owner == target:
        raise ActionNotExecutable("I cannot execute any action on the server owner!")
    
    if isinstance(target, discord.Member):
        if guild.me.top_role <= target.top_role:
            raise HierarchyException(target)
        
        if guild.owner == ctx.author:
            return
        
        if ctx.author.top_role <= target.top_role:
            raise HierarchyException(target, author_error=True)



import json
from typing import Tuple, Dict, Any

import discord
from redbot.core import commands # type: ignore

class RedBotException(discord.ClientException):
    """
    The base exception for the bot. 
    
    All other exceptions should inherit from this.
    """
    
    __slots__: Tuple[str, ...] = ()

class RedBotCommandError(commands.CommandError, RedBotException):
    """
    The base exception for the bot command errors.
    """
    
    __slots__: Tuple[str, ...] = ()
    
class HierarchyException(RedBotCommandError):
    """
    Raised when the bot is requested to perform an operation on a member
    that is higher than them in the guild hierarchy.
    """
    
    __slots__: Tuple[str, ...] = (
        "member",
        "author_error"
    )
    
    def __init__(
        self,
        member: discord.Member,
        *,
        author_error: bool = False
    ) -> None:
        self.member: discord.Member = member
        self.author_error: bool = author_error
        
        if author_error is False:
            super().__init__(
                f"**{member}**\'s top role is higher than mine. I can\'t do that!"
            )
        
        else:
            super().__init__(
                f"**{member}**\'s top role is higher than your top role. You can\'t do that!"
            )
    
class ActionNotExecutable(RedBotCommandError):
    """
    The exception for when an action is not executable.
    """
    
    def __init__(self, message):
        super().__init__("{}".format(message))
        
class EmbedGenException(Exception):
    @property
    def msg(self) -> str:
        return self._msg
    
    def __init__(
        self,
        msg="An exception was occured during EmbedGen operations.",
        *args,
        **kwargs
    ):
        self._msg = msg
        super().__init__(*args)
        
    def __str__(self) -> str:
        return self.msg
    
    def __repr__(self) -> str:
        return self.__str__()
    
class UnexpectedKwargsError(EmbedGenException):
    def __init__(
        self,
        unexpected_kwargs: Dict[str, Any],
        *args,
        **kwargs
    ):
        if "msg" in kwargs.keys():
            kwargs.pop("msg")
        
        self.kwargs = unexpected_kwargs
        
        msg = (
            "`EmbedGen.__init__()` caught unexpected keyword arguments! : "
            f"{json.dumps(obj=self.kwargs, indent=4, ensure_ascii=False)}"
        )
        super().__init__(*args, msg=msg, **kwargs)
        
class InvalidColorError(EmbedGenException):
    def __init__(
        self,
        invalid_color,
        *args,
        **kwargs
    ):
        self.color = invalid_color
        
        if "msg" in kwargs.keys():
            kwargs.pop("msg")
            
        super().__init__(
            *args,
            msg="Embed color must be an instance of `discord.Color`! : ",
            **kwargs
        )
        
class InvalidFieldError(EmbedGenException):
    def __init__(
        self,
        invalid_field,
        *args,
        **kwargs
    ):
        self.field = invalid_field
        
        if "msg" in kwargs.keys():
            kwargs.pop("msg")
            
        super().__init__(
            *args,
            msg="Embed field must have structure of `{'name': name, 'value': value}`",
            **kwargs
        )
        


import pathlib

from setuptools import setup

ROOT = pathlib.Path(__file__).parent

with open(ROOT / "README.md", "r", encoding="utf-8") as f:
    _long_description = f.read()
    
setup(
    name="melonutils",
    version="0.1.3",
    # if you want a cleaner versioning use the things below
    # setup_requires=['setuptools_scm'],
    # intall_requires=['setuptools_scm'],
    # use_scm_version={'write_to': 'melonutils/version.txt'},
    description="A collection of utilities for redbots, specially melon",
    long_description=_long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MelonBot-Development/melonutils",
    project_urls={
        "Code": "https://github.com/MelonBot-Development/melonutils",
        "Issue tracker": "https://github.com/MelonBot-Development/melonutils/issues"
    },
    author="Lemon Rose",
    author_email="support@melonbot.io",
    packages=["melonutils", "melonutils.core"],
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Topic :: Utilities",
    ],
    test_suite="tests",
    # include_package_data: to install data from MANIFEST.in
    include_package_data=True,
    zip_safe=False,
)



import re
from enum import Enum
from datetime import datetime
from abc import abstractmethod
from typing import Dict, Optional, List, Union, NoReturn, Any

from discord import Color

def validate_url(value) -> bool:
    return isinstance(value, str) and re.match("^https?", value)

@DeprecationWarning
class EmbedType(Enum):
    RICH = "rich"
    IMAGE = "image"
    VIDEO = "video"
    GIFV = "gifv"
    ARTICLE = "article"
    LINK = "link"

    def __contains__(self, item):
        """
        Override `in` keyword to check Embed`s type value.
        
        :param item:
        :return:
        """
        if type(item) == self.__class__:
            return super(EmbedType, self).__contains__(item)

        return str(item) in EmbedType.__members__.keys()

    @classmethod
    def from_value(cls, value: Union[str, EmbedType]) -> Union[EmbedType, NoReturn]: # type: ignore
        if type(value) == cls:
            return value

        elif type(value) == str:
            try:
                return EmbedType[value.upper()]
            except KeyError:
                raise KeyError("Unknown Embed Type {}".format(value))

        raise ValueError("EmbedType enum can be constructed only using string key or EmbedType object.")

class EmbedObject(object):
    """
    Represents property object used in discord`s embed structure.
    """

    def __repr__(self) -> str:
        return f"Embed.Object"

    @classmethod
    @abstractmethod
    def fromDict(cls, data: Dict[str, Any]):
        if not isinstance(data, dict):
            raise TypeError("Expected Dict[str, Any], caught {}".format(data.__class__))
        if cls == EmbedObject:
            raise NotImplementedError("Subclasses should implement the method!")

    @abstractmethod
    def toDict(self) -> dict:
        raise NotImplementedError("Subclasses should implement the method!")

@DeprecationWarning
class EmptyObject(EmbedObject):
    """
    Represents `empty` value in embed property.
    """

    def __init__(self, property_name: str, optional: bool = False) -> None:
        self.property_name = property_name
        self.optional = optional

    @classmethod
    def fromDict(cls, data: Dict[str, Union[str, bool]]) -> EmptyObject: # type: ignore
        return cls(
            property_name=data.get("property_name"),
            optional=data.get("optional") or True
        )

    def toDict(self) -> dict:
        raise ValueError("Empty object cannot be serialized into json.")

    def __str__(self) -> Union[str, NoReturn]:
        if not self.optional:
            raise ValueError(
                "Embed property '{}' cannot be empty!".format(self.property_name)
            )
        else:
            return "Embed property {} is empty.".format(self.property_name)

    def __repr__(self) -> str:
        return "Embed.Empty(property_name={},optional={})".format(self.property_name, self.optional)

    def __len__(self) -> int:
        """
        Tool to check whether property is empty or not.
        
        :return: Always 0, because this class always represents empty value.
        """
        return 0

    def __bool__(self) -> bool:
        """
        Tool to check whether property is empty or not.
        
        :return: Always False, because this class always represents empty value.
        """
        return False


class AuthorObject(EmbedObject):
    """
    Represents author objects on discord Embed.
    """

    def __init__(
        self, 
        name: str, 
        url: Optional[str] = None, 
        icon_url: Optional[str] = None,
        proxy_icon_url: Optional[str] = None
    ):
        if type(name) != str or len(name) > 256:
            raise ValueError("Author Object cannot have name longer than 256.")
        self.name = name

        if url is not None and validate_url(url):
            self.url = url
        else:
            raise ValueError("Invalid url!")

        if icon_url is not None and validate_url(icon_url):
            self.icon_url = icon_url
        else:
            raise ValueError("Invalid icon url!")

        if proxy_icon_url is not None and validate_url(proxy_icon_url):
            self.proxy_icon_url = proxy_icon_url
        else:
            raise ValueError("Invalid proxy icon url!")

    @classmethod
    def fromDict(cls, data: Union["AuthorObject", Dict[str, Any]]) -> AuthorObject: # type: ignore
        if not isinstance(data, dict):
            raise TypeError("Expected Dict[str, Any], caught {}".format(data.__class__))
        name = data.get("name")
        url = data.get("url") or None
        icon_url = data.get("icon_url") or None
        proxy_icon_url = data.get("proxy_icon_url") or None

        return cls(
            name=name,
            url=url,
            icon_url=icon_url,
            proxy_icon_url=proxy_icon_url
        )

    def toDict(self) -> Dict[str, str]:
        result = {
            "name": self.name
        }
        if self.url:
            result["url"] = self.url
        if self.icon_url:
            result["icon_url"] = self.icon_url
        if self.proxy_icon_url:
            result["proxy_icon_url"] = self.proxy_icon_url
        return result

    def __str__(self) -> str:
        return str(self.toDict())

    def __repr__(self) -> str:
        return ("Embed.Author(name={},url={},icon_url={},proxy_icon_url={})"
                .format(self.name, self.url, self.icon_url, self.proxy_icon_url))

class FooterObject(EmbedObject):
    """
    Represents footer objects on discord Embed.
    """
    
    def __init__(
        self, 
        text: Optional[str], 
        icon_url: Optional[str] = None,
        proxy_icon_url: Optional[str] = None
    ):
        self.text = text
        self.icon_url = icon_url if validate_url(icon_url) else None
        self.proxy_icon_url = proxy_icon_url if validate_url(proxy_icon_url) else None

        if proxy_icon_url is not None and validate_url(proxy_icon_url):
            self.proxy_icon_url = proxy_icon_url

    @classmethod
    def fromDict(
        cls, 
        data: Dict[str, Any]
    ) -> FooterObject: # type: ignore
        if not isinstance(data, dict):
            raise TypeError("Expected Dict[str, Any], caught {}".format(data.__class__))
        text = data.get("text")
        icon_url = data.get("icon_url") or None
        proxy_icon_url = data.get("proxy_icon_url") or None

        return cls(
            text=text,
            icon_url=icon_url,
            proxy_icon_url=proxy_icon_url
        )

    def toDict(self) -> Dict[str, str]:
        result = {
            "text": self.text
        }
        if self.icon_url:
            result["icon_url"] = self.icon_url
        if self.proxy_icon_url:
            result["proxy_icon_url"] = self.proxy_icon_url
        return result

    def __str__(self) -> str:
        return str(self.toDict())

    def __repr__(self) -> str:
        return ("Embed.Footer(text={},icon_url={},proxy_icon_url={})"
                .format(self.text, self.icon_url, self.proxy_icon_url))

class ImageObject(EmbedObject):
    """
    Represents image objects on discord Embed.
    Can be used at 'image', 'thumbnail' property (They share same options)
    """

    def __init__(
            self,
            url: str,
            proxy_url: Optional[str] = None,
            height: Optional[int] = None,
            width: Optional[int] = None
    ) -> None:
        if validate_url(url):
            self.url = url
        else:
            raise ValueError("Invalid url!")

        if validate_url(proxy_url):
            self.proxy_url: Optional[str] = proxy_url
        else:
            self.proxy_url: Optional[str] = None

        self.height = height
        self.width = width

    @classmethod
    def fromDict(
        cls, 
        data: Dict[str, Any]
    ) -> ImageObject: # type: ignore
        # Type Check
        if not isinstance(data, dict):
            raise TypeError("Expected Dict[str, Any], caught {}".format(data.__class__))
        url = data.get("url")
        proxy_url = data.get("proxy_url") or None
        height = data.get("height") or None
        width = data.get("width") or None
        return cls(url, proxy_url, height, width)

    def toDict(self) -> Dict[str, str]:
        result = {
            "url": self.url
        }
        if self.proxy_url is not None:
            result["proxy_url"] = self.proxy_url
        if self.height is not None:
            result["height"] = self.height
        if self.width is not None:
            result["width"] = self.width
        return result

    def __str__(self) -> str:
        return str(self.toDict())

    def __repr__(self) -> str:
        return ("Embed.Image(url={},proxy_url={},height={},width={}"
                .format(self.url, self.proxy_url, self.height, self.width))

ThumbnailObject = ImageObject


class VideoObject(EmbedObject):
    """
    Represents video objects on discord Embed.
    """

    def __init__(
        self, 
        url: str, 
        height: Optional[int] = None, 
        width: Optional[int] = None
    ):
        if validate_url(url):
            self.url = url
        else:
            raise ValueError("Invalid url!")

        self.height = height
        self.width = width

    @classmethod
    def fromDict(
        cls, 
        data: Dict[str, Any]
    ) -> VideoObject: # type: ignore
        # Type Check
        if not isinstance(data, dict):
            raise TypeError("Expected Dict[str, Any], caught {}".format(data.__class__))

        # Attribute Check
        url = data.get("url")
        height = data.get("height") or None
        width = data.get("width") or None
        return cls(
            url=url,
            height=height,
            width=width
        )

    def toDict(self) -> Dict[str, str]:
        result = {
            "url": self.url
        }
        if self.height is not None:
            result["height"] = self.height
        if self.width is not None:
            result["width"] = self.width
        return result

    def __str__(self) -> str:
        return str(self.toDict())

    def __repr__(self) -> str:
        return "Embed.Video(url={},height={},width={})".format(self.url, self.height, self.width)

class ProviderObject(EmbedObject):
    """
    Represents provider objects on discord Embed.
    """

    def __init__(
        self, 
        name: str, 
        url: str
    ) -> None:
        self.name = name
        self.url = url

    @classmethod
    def fromDict(
        cls, 
        data: Dict[str, Any]
    ) -> ProviderObject: # type: ignore
        # Type Check
        if not isinstance(data, dict):
            raise TypeError("Expected Dict[str, Any], caught {}".format(data.__class__))

        # Attribute Check
        try:
            name = data.get("name")
            url = data.get("url")
            return cls(
                name=name,
                url=url
            )

        except KeyError as e:
            raise ValueError(f"Invalid data is passed in VideoObject. : {data}. KeyError : {e}")

    def toDict(self) -> Dict[str, str]:
        result = {
            "name": self.name,
            "url": self.url
        }
        return result

    def __str__(self) -> str:
        return str(self.toDict())

    def __repr__(self) -> str:
        return "Embed.Provider(name={},url={})".format(self.name, self.url)

class Field(EmbedObject):
    """
    Represents field objects on discord Embed.
    """

    def __init__(
        self, 
        name: str, 
        value: str,
        inline: Optional[bool] = False
    ):

        if not Field.check_name(name):
            raise ValueError("")
        self.name = name
        if not Field.check_value(value):
            raise ValueError("")
        self.value = value
        if type(inline) != bool:
            inline = False
        self.inline = inline

    @classmethod
    def check_name(cls, name: str) -> bool:
        return type(name) is str and len(name) <= 256

    @classmethod
    def check_value(cls, value: str) -> bool:
        return type(value) is str and len(value) <= 1024

    @classmethod
    def fromDict(
        cls, 
        data: Union[Field, Dict[str, Union[str, bool]]] # type: ignore
    ) -> Optional[Field]: # type: ignore
        if isinstance(data, cls):
            return data
        if data is None or not isinstance(data, dict):
            raise TypeError("Expected Dict[str, Union[str, bool]], caught {}".format(data.__class__))
        try:
            name = data.get("name")
            if not cls.check_name(name):
                raise ValueError("")
            value = data.get("value")
            if not cls.check_value(value):
                raise ValueError("")
            inline = data.get("inline") or False
            if type(inline) != bool:
                raise TypeError("")
        # Does not except ValueError&TypeError, because it is intentionally raised to indicate error on given data.
        except KeyError as e:
            raise ValueError(f"Invalid data is passed in VideoObject. : {data}. KeyError : {e}")

    def toDict(self) -> Dict[str, str]:
        result = {
            "name": self.name,
            "value": self.value,
            "inline": self.inline
        }
        return result

    def __str__(self) -> str:
        return str(self.toDict())

    def __repr__(self) -> str:
        return ("Embed.Field(name={},value={},inline={})"
                .format(self.name, self.value, self.inline))

    def __getitem__(self, key) -> Optional[Any]:
        return getattr(self, key, None)

    def keys(self):
        def keyIter():
            yield "name"
            yield "value"
            yield "inline"
        return keyIter()

    def values(self):
        def valueIter():
            yield self.name
            yield self.value
            yield self.inline
        return valueIter()

    def items(self):
        def itemIter():
            yield "name", self.name
            yield "value", self.value
            yield "inline", self.inline
        return itemIter()

@NotImplemented
class Fields(EmbedObject, list):

    def __init__(self, fields: List[Field]):
        super().__init__()
        self.fields = []
        for field in fields:
            if not isinstance(field, Field):
                raise TypeError("field must be instance of Field object")
            self.fields.append(field)

    @classmethod
    def fromDict(cls, data: Dict[str, Any]) -> Fields: # type: ignore
        if isinstance(data, cls):
            return data
        if isinstance(data, list):
            return cls(data)

    def toDict(self) -> dict:
        return {
            "fields": str(self.fields)
        }

    def __getitem__(self, index: int) -> Field:
        return self.fields[index]

def check_title(value) -> bool:
    return type(value) == str and len(value) <= 256


def process_title(value: str) -> Union[str, NoReturn]:
    if check_title(value):
        return value

    raise ValueError("Embed title must be string object and its length must be lower than 256.")


def check_desc(value) -> bool:
    return type(value) == str and len(value) <= 2048


def process_desc(value: str) -> str:
    if check_desc(value):
        return value
    raise ValueError("Embed description must be string object and its length must be lower than 2048.")
